from typing import (
    TYPE_CHECKING,
    Optional,
    Union,
    cast,
    Dict,
    Any,
    Tuple,
    TypeVar,
)
from typing_extensions import Literal
from eth_keys import (
    keys,
)
from eth_keys.datatypes import (
    PrivateKey,
)
from eth_account.hdaccount import (
    key_from_seed,
    seed_from_mnemonic,
)
from eth_account.account import (
    Account as EthAccount,
)
from eth_account.datastructures import (
    SignedMessage,
    SignedTransaction,
)
from eth_account.messages import (
    SignableMessage,
)
from cfx_account.signers.local import LocalAccount
from eth_utils.crypto import (
    keccak,
)
from collections.abc import (
    Mapping,
)
from cytoolz import (
    dissoc,  # type: ignore
)
from hexbytes import (
    HexBytes,
)
from eth_utils.address import to_checksum_address
from eth_account.datastructures import (
    # SignedMessage,
    SignedTransaction,
)
from cfx_account._utils.signing import (
    sign_transaction_dict,
)
from cfx_account.transactions.legacy_transactions import (
    LegacyTransaction,
)
from cfx_address import (
    Base32Address,
    eth_eoa_address_to_cfx_hex,
)
from cfx_address.utils import (
    normalize_to,
)
from cfx_utils.types import (
    TxParam,
    TxDict,
    ChecksumAddress,
    HexStr,
)
from cfx_utils.decorators import (
    combomethod,
)
from cfx_account.types import (
    KeyfileDict,
)

if TYPE_CHECKING:
    from conflux_web3 import Web3

CONFLUX_DEFAULT_PATH = "m/44'/503'/0'/0/0"
VRS = TypeVar("VRS", bytes, HexStr, int)


class Account(EthAccount):

    # _default_network_id: Optional[int]=None
    w3: Optional["Web3"] = None

    _use_unaudited_hdwallet_features = True

    @combomethod
    def set_w3(self, w3: "Web3") -> None:
        self.w3 = w3

    # def set_default_network_id(self, network_id: int):
    #     self._default_network_id = network_id

    @combomethod
    def from_key(
        self,
        private_key: Union[bytes, str, PrivateKey],
        network_id: Optional[int] = None,
    ) -> LocalAccount:
        """
        returns a LocalAccount object

        :param str private_key: the raw private key
        :param Optional[int] network_id: target network of the account, defaults to None
        :return LocalAccount: object with methods for signing and encrypting

        >>> acct = Account.from_key(
        ... 0xb25c7db31feed9122727bf0939dc769a96564b2de4c4726d035b36ecf1e5b364)
        >>> acct.address
        '0x1ce9454909639d2d17a3f753ce7d93fa0b9ab12e'
        >>> acct.key
        HexBytes('0xb25c7db31feed9122727bf0939dc769a96564b2de4c4726d035b36ecf1e5b364')

        # These methods are also available: sign_message(), sign_transaction(), encrypt()
        # They correspond to the same-named methods in Account.*
        # but without the private key argument
        """
        key = self._parse_private_key(private_key)
        return LocalAccount(
            key,
            self,
            # use network_id is it is not None
            # then use None if self.w3 is not set
            # if self.w3 is set, use self.w3.cfx.chain_id
            network_id or (self.w3 and self.w3.cfx.chain_id),
        )

    @combomethod
    def sign_transaction(
        self, transaction_dict: TxParam, private_key: Union[bytes, str, PrivateKey],
        blobs: Optional[Any] = None,
    ) -> SignedTransaction:
        """
        Sign a transaction using a local private key. Produces signature details
        and the hex-encoded transaction suitable for broadcast using
        :meth:`w3.cfx.send_raw_transaction() <conflux_web3.client.ConfluxClient.send_raw_transaction>`.

        Refer to `Interact with a Contract
        <https://python-conflux-sdk.readthedocs.io/en/latest/examples/10-send_raw_transaction.html#interact-with-a-contract>`_
        to see how to sign for a contract method using `build_transaction`

        :param TxParam transaction_dict: the transaction with keys:
          nonce, chainId, to, data, value, storageLimit, epochHeight, gas, and gasPrice.
        :param Union[bytes,str,PrivateKey] private_key: private_key to be used for signing
        :raises TypeError: transaction_dict is not a dict-like object
        :raises ValueError: transaction's from field does not match private_key
        :return SignedTransaction: an attribute dict contains various details about the signature - most
          importantly the fields: v, r, and s

        >>> transaction = {
                # Note that the address must be in Base32 format or native bytes:
                'to': 'cfxtest:aak7fsws4u4yf38fk870218p1h3gxut3ku00u1k1da',
                'nonce': 1,
                'value': 1,
                'gas': 100,
                'gasPrice': 1,
                'storageLimit': 100,
                'epochHeight': 100,
                'chainId': 1
            }
        >>> key = '0xcc7939276283a32f60d2fad7d16cac972300308fe99ec98d0e63765d02e24863'
        >>> signed = Account.sign_transaction(transaction, key)
        {'hash': HexBytes('0x692a0ea530a264f4e80ce39f393233e90638ef929c8706802e15299fd0b042b9'),
            'r': 74715349327018893060702835194036838027583623083228589573427622179540208747230,
            'raw_transaction': HexBytes('0xf861dd0101649413d2ba4ed43542e7c54fbb6c5fccb9f269c1f94c016464018080a0a52f639cbed11262a7b88d0a37aef909aa7dc2c36c40689a3d52b8bd1d9482dea054f3bdeb654f73704db4cbc12451fb4c9830ef62b0f24de1a40e4b6fe10f57b2'),  # noqa: E501
            's': 38424933894051759888751352802050752143518665905311311986258635963723328477106,
            'v': 0}
        >>> w3.cfx.sendRawTransaction(signed.raw_transaction)
        """
        if not isinstance(transaction_dict, Mapping):
            raise TypeError(
                "transaction_dict must be dict-like, got %r" % transaction_dict
            )

        account: LocalAccount = self.from_key(private_key)
        transaction_dict = cast(TxDict, transaction_dict)
        # allow from field, *only* if it matches the private key
        if "from" in transaction_dict:
            if normalize_to(transaction_dict["from"], None) == normalize_to(
                account.address, None
            ):
                sanitized_transaction = cast(TxDict, dissoc(transaction_dict, "from"))
            else:
                raise ValueError(
                    "transaction[from] does match key's hex address: "
                    f"from's hex address is {Base32Address(transaction_dict['from']).hex_address}, "
                    f"key's hex address is {account.hex_address}"
                )

        else:
            sanitized_transaction = transaction_dict

        # sign transaction
        (
            v,
            r,
            s,
            raw_transaction,
        ) = sign_transaction_dict(
            account._key_obj, sanitized_transaction
        )  # type: ignore

        transaction_hash = keccak(raw_transaction)

        return SignedTransaction(
            raw_transaction=HexBytes(raw_transaction),
            hash=HexBytes(transaction_hash),
            r=r,
            s=s,
            v=v,
        )

    @combomethod
    def from_mnemonic(
        self,
        mnemonic: str,
        passphrase: str = "",
        account_path: str = CONFLUX_DEFAULT_PATH,
        network_id: Optional[int] = None,
    ) -> LocalAccount:
        """
        Generate an account from a mnemonic.

        :param str mnemonic: space-separated list of BIP39 mnemonic seed words
        :param str passphrase: Optional passphrase used to encrypt the mnemonic, defaults to ""
        :param str account_path: pecify an alternate HD path for deriving the seed using
            BIP32 HD wallet key derivation, defaults to CONFLUX_DEFAULT_PATH(m/44'/503'/0'/0/0)
        :param Optional[int] network_id: the network id of returned account, defaults to None
        :return LocalAccount: a LocalAccount object

        :examples:

        >>> from cfx_account import Account
        >>> acct = Account.from_mnemonic('faint also eye industry survey unhappy boil public lemon myself cube sense', network_id=1)
        >>> acct.address
        'cfxtest:aargrnff46pmuy2g1mmrntctkhr5mzamh6nmg361n0'
        """
        # acct: LocalAccount = super().from_mnemonic(mnemonic, passphrase, account_path)
        seed = seed_from_mnemonic(mnemonic, passphrase)
        private_key = key_from_seed(seed, account_path)
        key = self._parse_private_key(private_key)
        return LocalAccount(key, self, network_id or (self.w3 and self.w3.cfx.chain_id))

    @combomethod
    def create_with_mnemonic(
        self,
        passphrase: str = "",
        num_words: int = 12,
        language: str = "english",
        account_path: str = CONFLUX_DEFAULT_PATH,
        network_id: Optional[int] = None,
    ) -> Tuple[LocalAccount, str]:
        """
        Create mnemonic and derive an account using parameters

        :param str passphrase: Extra passphrase to encrypt the seed phrase, defaults to ""
        :param int num_words: Number of words to use with seed phrase. Default is 12 words.
                              Must be one of [12, 15, 18, 21, 24].
        :param str language: Language to use for BIP39 mnemonic seed phrase, defaults to "english"
        :param str account_path: Specify an alternate HD path for deriving the seed using
            BIP32 HD wallet key derivation, defaults to CONFLUX_DEFAULT_PATH(m/44'/503'/0'/0/0)
        :param Optional[int] network_id: the network id of returned account, defaults to None
        :return Tuple[LocalAccount, str]: a LocalAccount object and related mnemonic
        """
        acct, mnemonic = super().create_with_mnemonic(
            passphrase, num_words, language, account_path
        )
        if network_id is not None:
            acct.network_id = network_id
        return acct, mnemonic

    @classmethod
    def encrypt(  # type: ignore
        cls,
        private_key: Union[bytes, str, PrivateKey],
        password: str,
        kdf: Optional[Literal["scrypt", "pbkdf2"]] = None,
        iterations: Optional[int] = None,
    ) -> KeyfileDict:
        return super().encrypt(private_key, password, kdf, iterations)  # type: ignore

    @staticmethod
    def decrypt(
        keyfile_json: Union[Dict[str, Any], str, KeyfileDict], password: str
    ) -> HexBytes:
        """
        Decrypts a keyfile and returns the secret key.

        :param Union[Dict[str,Any],str,KeyfileDict] keyfile_json: encrypted keyfile
        :param str password: the password that was used to encrypt the key
        :return HexBytes: the hex private key
        """
        return EthAccount.decrypt(keyfile_json, password)

    @combomethod
    def recover_transaction(
        self, serialized_transaction: Union[bytes, HexStr, str]
    ) -> ChecksumAddress:
        """
        Get the address of the account that signed this transaction.

        :param Union[bytes,HexStr,str] serialized_transaction: the complete signed transaction
        :return ChecksumAddress: address of signer, hex-encoded & checksummed

        :example:

        >>> raw_transaction = '0xf86a8086d55698372431831e848094f0109fc8df283027b6285cc889f5aa624eac1f55843b9aca008025a009ebb6ca057a0535d6186462bc0b465b561c94a295bdb0621fc19208ab149a9ca0440ffd775ce91a833ab410777204d5341a6f9fa91216a6f3ee2c051fea6a0428'  # noqa: E501
        >>> Account.recover_transaction(raw_transaction)
        '0x1c7536e3605d9c16a7a3d7b1898e529396a65c23'
        """
        txn_bytes = HexBytes(serialized_transaction)
        # TODO: replace with Transaction.from_bytes()
        txn = LegacyTransaction.from_bytes(txn_bytes)
        recovered_address = self._recover_hash(txn.hash(), vrs=txn.vrs())  # type: ignore
        return to_checksum_address(eth_eoa_address_to_cfx_hex(recovered_address))

    @combomethod
    def create(
        self, extra_entropy: str = "", network_id: Optional[int] = None
    ) -> LocalAccount:
        """
        Creates a new private key, and returns it as a :class:`~cfx_account.signers.local.LocalAccount`.

        :param str extra_entropy: Add extra randomness to the randomness provided by your OS, defaults to ''
        :param Optional[int] network_id: the network id of the generated account, which determines the address encoding, defaults to None
        :return LocalAccount: an object with private key and convenience methods

        :examples:

        >>> from cfx_account import Account
        >>> acct = Account.create()
        >>> acct.address
        '0x187EE3Cb948fFb5a34417344f7fA01c638aa2F96'
        >>> Account.create(network_id=1029).address
        'cfx:aapapvutg83zauuexwdrgjp4v6xm3dkbm2vevh1rub'
        """
        acct: LocalAccount = super().create(extra_entropy)
        if network_id is not None:
            acct.network_id = network_id
        return acct

    @combomethod
    def sign_message(
        self,
        signable_message: SignableMessage,
        private_key: Union[bytes, HexStr, int, keys.PrivateKey],
    ) -> SignedMessage:
        """
        Sign the provided encoded message. The message is encoded according to CIP-23_

        :param SignableMessage signable_message: an encoded message generated by `encode_defunct` or `encode_structured_data`
        :param Union[bytes,HexStr,int,keys.PrivateKey] private_key: the private key used to sign message
        :return SignedMessage: a signed message object

        :examples:

        >>> from from cfx_account import Account
        >>> from cfx_account.messages import encode_structured_data, encode_defunct
        >>> encoded_message = encode_defunct(text=message)
        >>> acct = Account.create()
        >>> signed = acct.sign_message(encoded_message)
        >>> signed.signature
        '0xd72ea2020802d6dfce0d49fc1d92a16b43baa58fc152d6f437d852a014e0c5740b3563375b0b844a835be4f1521b4ae2a691048622f70026e0470acc5351043a01'
        >>> assert acct.hex_address == Account.recover_message(encoded_message, signature=signed.signature)

        >>> # https://github.com/Conflux-Chain/CIPs/blob/master/CIPs/cip-23.md#typed-data
        >>> typed_data = { "types": { "CIP23Domain": [ ... ] }, ... }
        >>> encoded_data = encode_structured_data(typed_data)
        >>> signed = acct.sign_message(encoded_data)
        >>> signed.signature
        '0xd7fb6dca3b084ae3a9bf1ea3527de7a9bc2bd40e0c38d3faf9da214f1d5637ab2944a8a993dc59365c1e74e18a1589b358e3fb81bd03892d159f221e8ac765c701'
        >>> assert acct.hex_address == Account.recover_message(encoded_data, signature=signed.signature)

        .. _CIP-23: https://github.com/Conflux-Chain/CIPs/blob/master/CIPs/cip-23.md
        """
        return super().sign_message(signable_message, private_key)

    @combomethod
    def recover_message(
        self,
        signable_message: SignableMessage,
        vrs: Optional[Tuple[VRS, VRS, VRS]] = None,
        signature: Optional[bytes] = None,
    ) -> ChecksumAddress:
        """
        Get the address of the account that signed the given message.
        Must specify exactly one of vrs or signature.
        Refer to :meth:`~cfx_account.account.Account.sign_message` for usage examples.

        :param SignableMessage signable_message: an encoded message generated by `encode_defunct` or `encode_structured_data`
        :param Optional[Tuple[VRS,VRS,VRS]] vrs: the three pieces generated by an elliptic curve signature, defaults to None
        :param Optional[bytes] signature: signature bytes concatenated as r+s+v, defaults to None
        :return ChecksumAddress: the checksum address of the account that signed the given message
        """
        recovered_address = super().recover_message(signable_message, vrs, signature)
        return to_checksum_address(eth_eoa_address_to_cfx_hex(recovered_address))



import json, pytest
from hexbytes import HexBytes

from eth_utils.exceptions import (
    ValidationError,
)
from cfx_account import Account
from cfx_account.messages import encode_structured_data, encode_defunct

private_key = "0x0123456789abcdef0123456789abcdef0123456789abcdef0123456789abcdef"
typed_data = json.load(open("tests/typed-data.json"))
typed_data_without_chainId = json.load(open("tests/typed-data-without-chainid.json"))
message = "Hello World"

# signature from js-conflux-sdk
message_signature = "0xd72ea2020802d6dfce0d49fc1d92a16b43baa58fc152d6f437d852a014e0c5740b3563375b0b844a835be4f1521b4ae2a691048622f70026e0470acc5351043a01"
typed_data_signature = "0xd7fb6dca3b084ae3a9bf1ea3527de7a9bc2bd40e0c38d3faf9da214f1d5637ab2944a8a993dc59365c1e74e18a1589b358e3fb81bd03892d159f221e8ac765c701"

def test_encode_defunct():
    encoded_message = encode_defunct(text=message)
    address = Account.recover_message(encoded_message, signature=HexBytes(message_signature))
    assert address == Account.from_key(private_key).address

def test_encode_structured_data():
    encoded_data = encode_structured_data(typed_data)
    address = Account.recover_message(encoded_data, signature=HexBytes(typed_data_signature))
    assert address == Account.from_key(private_key).address

def test_sign_message():
    encoded_message = encode_defunct(text=message)
    acct = Account.from_key(private_key)
    signed = acct.sign_message(encoded_message)
    address = Account.recover_message(encoded_message, signature=signed.signature)
    assert address == acct.address

def test_sign_structured_data():
    encoded_data = encode_structured_data(typed_data)
    acct = Account.from_key(private_key)
    signed = acct.sign_message(encoded_data)
    address = Account.recover_message(encoded_data, signature=signed.signature)
    assert address == acct.address

def test_sign_structured_data_without_chainId():
    with pytest.raises(ValidationError):
        encode_structured_data(typed_data_without_chainId)



from typing import Any, Callable, ClassVar, Dict, Tuple, cast, Union

import rlp
from cfx_utils.types import TxParam
from cytoolz import dissoc  # type: ignore
from cytoolz import merge  # type: ignore
from cytoolz import pipe  # type: ignore
from cytoolz import partial
from eth_account._utils.legacy_transactions import TRANSACTION_DEFAULTS
from eth_account._utils.validation import is_int_or_prefixed_hexstr
from eth_rlp import HashableRLP
from eth_utils.curried import apply_formatters_to_dict
from hexbytes import HexBytes
from rlp.sedes import Binary, big_endian_int, binary

from cfx_account.transactions.transaction_utils import (
    LEGACY_TRANSACTION_FORMATTERS,
    is_empty_or_valid_base32_address,
)

from .base import TransactionImplementation  # type: ignore

LEGACY_UNSIGNED_TRANSACTION_FIELDS = (
    ("nonce", big_endian_int),
    ("gasPrice", big_endian_int),
    ("gas", big_endian_int),
    ("to", Binary.fixed_length(20, allow_empty=True)),
    ("value", big_endian_int),
    ("storageLimit", big_endian_int),
    ("epochHeight", big_endian_int),
    ("chainId", big_endian_int),
    ("data", binary),
)


class UnsignedLegacyTransactionImpl(HashableRLP):
    fields = LEGACY_UNSIGNED_TRANSACTION_FIELDS


class LegacyTransactionImpl(HashableRLP):
    fields = (
        ("tx_meta", UnsignedLegacyTransactionImpl),
        ("v", big_endian_int),
        ("r", big_endian_int),
        ("s", big_endian_int),
    )


class LegacyTransaction(TransactionImplementation):

    # ImplType: Type[HashableRLP]
    impl: Union[LegacyTransactionImpl, UnsignedLegacyTransactionImpl]

    transaction_type: ClassVar[int] = 0

    def __init__(self, tx_dict: TxParam):
        if "type" in tx_dict:
            tx_dict.pop("type")  # type: ignore

        # signed
        if "v" in tx_dict:
            self.ImplType = LegacyTransactionImpl
            chain_naive_transaction = dissoc(tx_dict, "v", "r", "s")
            self.impl = LegacyTransactionImpl(
                tx_meta=UnsignedLegacyTransactionImpl(**chain_naive_transaction),
                v=tx_dict["v"],
                r=tx_dict["r"],
                s=tx_dict["s"],
            )
            # self.impl = LegacyTransaction(tx_dict)
        else:
            # Unsigned
            self.ImplType = UnsignedLegacyTransactionImpl
            self.impl = serializable_unsigned_transaction_from_dict(tx_dict)

    def hash(self) -> bytes:
        if self.ImplType is UnsignedLegacyTransactionImpl:
            return self.impl.hash()
        else:
            return self.impl[0].hash() # type: ignore

    def from_dict(self, tx_dict: TxParam) -> "LegacyTransaction":
        return LegacyTransaction(tx_dict)

    def as_dict(self) -> Dict[str, Any]:
        if self.ImplType is UnsignedLegacyTransactionImpl:
            return self.impl.as_dict()  # type: ignore
        else:
            vrs = self.vrs()
            return {
                **self.impl.as_dict()[0].as_dict(),  # type: ignore
                "v": vrs[0],
                "r": vrs[1],
                "s": vrs[2],
            }  # type: ignore

    def append_signature(self, *, v: int, r: int, s: int) -> "LegacyTransaction":
        if self.is_signed():
            raise ValueError("Transaction is already signed")
        else:
            self.ImplType = LegacyTransactionImpl
            self.impl = LegacyTransactionImpl(tx_meta=self.impl, v=v, r=r, s=s)
        return self

    def is_signed(self) -> bool:
        return self.ImplType is LegacyTransactionImpl

    def encode(self, *, allow_unsigned: bool = False) -> bytes:
        if (not self.is_signed()) and (not allow_unsigned):
                raise ValueError("Transaction is not signed")
       
        return rlp.encode(self.impl)

    def vrs(self) -> Tuple[int, int, int]:
        if self.ImplType is LegacyTransactionImpl:
            return vrs_from(self.impl)
        else:
            raise ValueError("Unsigned transaction does not have vrs")

    @classmethod
    def from_bytes(cls, encoded_transaction: HexBytes) -> "LegacyTransaction":
        impl = cast(
            LegacyTransactionImpl, LegacyTransactionImpl.from_bytes(encoded_transaction)
        )
        return LegacyTransaction(
            {
                **impl[0].as_dict(), # type: ignore
                "v": impl.v,
                "r": impl.r,
                "s": impl.s,
            }
        )


def serializable_unsigned_transaction_from_dict(
    transaction_dict: TxParam,
) -> UnsignedLegacyTransactionImpl:
    assert_valid_fields(transaction_dict)
    filled_transaction = pipe(
        transaction_dict,
        dict,
        partial(merge, TRANSACTION_DEFAULTS),
        apply_formatters_to_dict(LEGACY_TRANSACTION_FORMATTERS),
    )
    serializer = UnsignedLegacyTransactionImpl
    return serializer.from_dict(filled_transaction)


def encode_transaction(
    unsigned_transaction: UnsignedLegacyTransactionImpl, vrs: Tuple[int, int, int]
) -> bytes:
    (v, r, s) = vrs
    chain_naive_transaction = dissoc(unsigned_transaction.as_dict(), "v", "r", "s")
    signed_transaction = LegacyTransactionImpl(
        tx_meta=UnsignedLegacyTransactionImpl(**chain_naive_transaction), v=v, r=r, s=s
    )
    return rlp.encode(signed_transaction)  # type: ignore


TRANSACTION_VALID_VALUES: Dict[str, Callable[[Any], bool]] = {
    "nonce": is_int_or_prefixed_hexstr,
    "gasPrice": is_int_or_prefixed_hexstr,
    "gas": is_int_or_prefixed_hexstr,
    "to": is_empty_or_valid_base32_address,
    "value": is_int_or_prefixed_hexstr,
    "storageLimit": is_int_or_prefixed_hexstr,
    "epochHeight": is_int_or_prefixed_hexstr,
    "chainId": is_int_or_prefixed_hexstr,
    "data": lambda val: isinstance(val, (int, str, bytes, bytearray)),  # type: ignore
}

ALLOWED_TRANSACTION_KEYS = {
    "nonce",
    "gasPrice",
    "gas",
    "to",
    "value",
    "storageLimit",
    "epochHeight",
    "chainId",
    "data",
}

REQUIRED_TRANSACITON_KEYS = ALLOWED_TRANSACTION_KEYS.difference(
    TRANSACTION_DEFAULTS.keys()
)


def assert_valid_fields(transaction_dict: Any) -> None:
    # check if any keys are missing
    missing_keys = REQUIRED_TRANSACITON_KEYS.difference(transaction_dict.keys())
    if missing_keys:
        raise TypeError("Transaction must include these fields: %r" % missing_keys)

    # check if any extra keys were specified
    superfluous_keys = set(transaction_dict.keys()).difference(ALLOWED_TRANSACTION_KEYS)
    if superfluous_keys:
        raise TypeError(
            "Transaction must not include unrecognized fields: %r" % superfluous_keys
        )

    # check for valid types in each field
    valid_fields: Dict[str, bool]
    valid_fields = apply_formatters_to_dict(TRANSACTION_VALID_VALUES, transaction_dict)
    if not all(valid_fields.values()):
        invalid = {
            key: transaction_dict[key]
            for key, valid in valid_fields.items()
            if not valid
        }
        raise TypeError("Transaction had invalid fields: %r" % invalid)


def vrs_from(transaction: LegacyTransactionImpl) -> Tuple[int, int, int]:
    return (transaction.v, transaction.r, transaction.s)  # type: ignore



from typing import Any, Dict

from cfx_address import Base32Address
from cfx_utils.types import TxDict, TxParam
from eth_account._utils.validation import VALID_EMPTY_ADDRESSES, is_none
from eth_utils.conversions import to_bytes, to_int
from eth_utils.curried import (
    apply_formatter_to_array,
    apply_formatters_to_dict,
    apply_one_of_formatters,
    hexstr_if_str,
)
from eth_utils.types import is_bytes, is_string
from hexbytes import HexBytes
from rlp.sedes import BigEndianInt, Binary, CountableList
from rlp.sedes import List as ListSedesClass
from toolz import assoc, identity, merge


def is_empty_or_valid_base32_address(val: Any) -> bool:
    if val in VALID_EMPTY_ADDRESSES:
        return True
    else:
        return Base32Address.is_valid_base32(val)


LEGACY_TRANSACTION_FORMATTERS = {
    "nonce": hexstr_if_str(to_int),
    "gasPrice": hexstr_if_str(to_int),
    "gas": hexstr_if_str(to_int),
    "to": apply_one_of_formatters(
        (
            (Base32Address.is_valid_base32, lambda val: HexBytes(Base32Address(val).hex_address)),  # type: ignore
            (is_string, hexstr_if_str(to_bytes)),
            (is_bytes, identity),
            (is_none, lambda val: b""),  # type: ignore
        )
    ),
    "value": hexstr_if_str(to_int),
    "storageLimit": hexstr_if_str(to_int),
    "epochHeight": hexstr_if_str(to_int),
    "chainId": hexstr_if_str(to_int),
    "data": hexstr_if_str(to_bytes),
    "v": hexstr_if_str(to_int),
    "r": hexstr_if_str(to_int),
    "s": hexstr_if_str(to_int),
}

TYPED_TRANSACTION_FORMATTERS = merge(
    LEGACY_TRANSACTION_FORMATTERS,
    {
        "chainId": hexstr_if_str(to_int),
        "type": hexstr_if_str(to_int),
        "accessList": apply_formatter_to_array(
            apply_formatters_to_dict(
                {
                    "address": apply_one_of_formatters(
                        (
                            (Base32Address.is_valid_base32, lambda val: HexBytes(Base32Address(val).hex_address)),  # type: ignore
                            (is_string, hexstr_if_str(to_bytes)),
                            (is_bytes, identity),
                        )
                    ),
                    "storageKeys": apply_formatter_to_array(hexstr_if_str(to_int)),
                }
            ),
        ),
        "maxPriorityFeePerGas": hexstr_if_str(to_int),
        "maxFeePerGas": hexstr_if_str(to_int),
        "maxFeePerBlobGas": hexstr_if_str(to_int),
        "blobVersionedHashes": apply_formatter_to_array(hexstr_if_str(to_bytes)),
    },
)
access_list_sede_type = CountableList(
    ListSedesClass(
        [
            Binary.fixed_length(20, allow_empty=False),
            CountableList(BigEndianInt(32)),
        ]
    ),
)


# returns a copy of the transaction dict with the 'type' field converted to int
def copy_ensuring_int_transaction_type(transaction_dict: TxParam) -> Dict[str, Any]:
    if "type" not in transaction_dict:
        if "gasPrice" in transaction_dict:
            if "accessList" in transaction_dict:
                # access list txn - type 1
                return assoc(transaction_dict, "type", 1)
            else:
                return assoc(transaction_dict, "type", 0)
        # elif any(
        #     type_2_arg in transaction_dict
        #     for type_2_arg in ("maxFeePerGas", "maxPriorityFeePerGas")
        # ):
        else:
            return assoc(transaction_dict, "type", 2)
    cpy = transaction_dict.copy()
    if isinstance(cpy["type"], str):
        cpy["type"] = int(cpy["type"], 16)
    return cpy



from cfx_address import Base32Address
from cfx_account import Account
from cfx_account.transactions.cip1559_transactions import CIP1559Transaction
from tests.test_utils import assert_hex_equal

signed_cip1559_transaction_dict = {
    "type": 2,
    "nonce": 100,
    "maxPriorityFeePerGas": 100,
    "maxFeePerGas": 100,
    "gas": 100,
    "to": Base32Address("0x19578CF3c71eaB48cF810c78B5175d5c9E6Ef441", 10),
    "value": 100,
    "data": "Hello, World".encode('utf-8'),
    "storageLimit": 100,
    "epochHeight": 100,
    "chainId": 100,
    "accessList": [
        {
            "address": Base32Address("0x19578CF3c71eaB48cF810c78B5175d5c9E6Ef441", 10),
            "storageKeys": [
                "0x1234567890abcdef1234567890abcdef1234567890abcdef1234567890abcdef"
            ],
        }
    ],
    "r": 1,
    "s": 1,
    "v": 0,
    # "hash_without_sig": "0x3da56dbe2b76c41135c2429f3035cd79b1abb68902cf588075c30d4912e71cf3",
    # "raw": "63667802f869f864646464649419578cf3c71eab48cf810c78b5175d5c9e6ef441646464648c48656c6c6f2c20576f726c64f838f79419578cf3c71eab48cf810c78b5175d5c9e6ef441e1a01234567890abcdef1234567890abcdef1234567890abcdef1234567890abcdef800101",
}

unsigned_cip1559_transaction_dict = {
    "type": 2,
    "nonce": 100,
    "maxPriorityFeePerGas": 100,
    "maxFeePerGas": 100,
    "gas": 100,
    "to": Base32Address("0x19578CF3c71eaB48cF810c78B5175d5c9E6Ef441", 10),
    "value": 100,
    "data": "Hello, World".encode('utf-8'),
    "storageLimit": 100,
    "epochHeight": 100,
    "chainId": 100,
    "accessList": [
        {
            "address": Base32Address("0x19578CF3c71eaB48cF810c78B5175d5c9E6Ef441", 10),
            "storageKeys": [
                "0x1234567890abcdef1234567890abcdef1234567890abcdef1234567890abcdef"
            ],
        }
    ]
    # "hash_without_sig": "0x3da56dbe2b76c41135c2429f3035cd79b1abb68902cf588075c30d4912e71cf3",
    # "raw": "63667802f869f864646464649419578cf3c71eab48cf810c78b5175d5c9e6ef441646464648c48656c6c6f2c20576f726c64f838f79419578cf3c71eab48cf810c78b5175d5c9e6ef441e1a01234567890abcdef1234567890abcdef1234567890abcdef1234567890abcdef800101",
}


def test_cip1559_transaction_hashing_and_encoding():
    cip1559_transaction = CIP1559Transaction(signed_cip1559_transaction_dict)
    assert_hex_equal(
        cip1559_transaction.hash(),
        "3da56dbe2b76c41135c2429f3035cd79b1abb68902cf588075c30d4912e71cf3"
    )
    assert_hex_equal(
        cip1559_transaction.encode(),
        "63667802f869f864646464649419578cf3c71eab48cf810c78b5175d5c9e6ef441646464648c48656c6c6f2c20576f726c64f838f79419578cf3c71eab48cf810c78b5175d5c9e6ef441e1a01234567890abcdef1234567890abcdef1234567890abcdef1234567890abcdef800101"
    )

def test_signing_cip1559_transaction():
    acct = Account.create()
    raw_tx = acct.sign_transaction(unsigned_cip1559_transaction_dict).raw_transaction
    assert raw_tx



from abc import ABC, abstractmethod
from typing import Any, Dict, Tuple
from typing_extensions import Self
from cfx_utils.types import TxParam

class TransactionImplementation(ABC):
    """
    Abstract class that every type of transaction must implement.
    Should not be imported or used by clients of the library.
    """
    
    transaction_type: int

    # blob_data: Optional[BlobPooledTransactionData] = None
    
    @abstractmethod
    def __init__(self, transaction_dict: TxParam):
        ...

    @abstractmethod
    def hash(self) -> bytes:
        pass

    @abstractmethod
    def as_dict(self) -> Dict[str, Any]:
        pass

    @abstractmethod
    def vrs(self) -> Tuple[int, int, int]:
        pass
    
    @abstractmethod
    def is_signed(self) -> bool:
        pass
    
    @abstractmethod
    def append_signature(self, *, v: int, r: int, s: int) -> Self:
        pass
    
    @abstractmethod
    def encode(self, *, allow_unsigned: bool = False) -> bytes:
        pass



from typing import (
    Dict,
    Any,
)

from eth_utils.exceptions import (
    ValidationError,
)
from cfx_account._utils.structured_data.eth_account_legacy_validation import (
    used_header_fields,
    validate_types_attribute,
    validate_primaryType_attribute,
    validate_has_attribute,
    validate_field_declared_only_once_in_struct,
)

def validate_CIP23Domain_schema(structured_data: Dict[str, Any]):
    # Check that the `types` attribute contains `CIP23Domain` schema declaration
    if "CIP23Domain" not in structured_data["types"]:
        raise ValidationError("`CIP23Domain struct` not found in types attribute")
    # Check that the names and types in `CIP23Domain` are what are mentioned in the CIP-23
    # and they are declared only once (if defined at all)
    CIP23Domain_data = structured_data["types"]["CIP23Domain"]
    header_fields = used_header_fields(CIP23Domain_data)
    # CIP-23 requires `chainId` field
    if "chainId" not in header_fields:
        raise ValidationError(f"Field 'chainId' must be defined in {structured_data}")
    for field in header_fields:
        validate_field_declared_only_once_in_struct(field, CIP23Domain_data, "CIP23Domain")


def validate_structured_data(structured_data: Dict[str, Any]):
    # validate the `types` attribute
    validate_types_attribute(structured_data)
    # validate the `CIP23Domain` struct of `types` attribute
    validate_CIP23Domain_schema(structured_data)
    # validate the `primaryType` attribute
    validate_primaryType_attribute(structured_data)
    # Check that there is a `domain` attribute in the structured data
    validate_has_attribute("domain", structured_data)
    # Check that there is a `message` attribute in the structured data
    validate_has_attribute("message", structured_data)



from typing import (
    Tuple,
)
from cfx_utils.types import TxParam
from eth_keys.datatypes import PrivateKey

from cfx_utils.token_unit import (
    to_int_if_drip_units,
)

from ..transactions.transactions import (
    Transaction,
)


def sign_transaction_dict(
    eth_key: PrivateKey, transaction_dict: TxParam
) -> Tuple[int, int, int, bytes]:
    if not transaction_dict.get("gasPrice") is None:
        transaction_dict["gasPrice"] = to_int_if_drip_units(transaction_dict["gasPrice"])  # type: ignore
    if not transaction_dict.get("value") is None:
        transaction_dict["value"] = to_int_if_drip_units(transaction_dict["value"])  # type: ignore
    if not transaction_dict.get("maxFeePerGas") is None:
        transaction_dict["maxFeePerGas"] = to_int_if_drip_units(transaction_dict["maxFeePerGas"])  # type: ignore
    if not transaction_dict.get("maxPriorityFeePerGas") is None:
        transaction_dict["maxPriorityFeePerGas"] = to_int_if_drip_units(transaction_dict["maxPriorityFeePerGas"])  # type: ignore
    # generate RLP-serializable transaction, with defaults filled
    transaction = Transaction.from_dict(transaction_dict)

    transaction_hash = transaction.hash()

    # sign with private key
    (v, r, s) = sign_transaction_hash(eth_key, transaction_hash)

    # serialize transaction with rlp
    raw_transaction = transaction.append_signature(v=v,r=r,s=s).encode()

    return (v, r, s, raw_transaction)


def sign_transaction_hash(
    key: PrivateKey, transaction_hash: bytes
) -> Tuple[int, int, int]:
    signature = key.sign_msg_hash(transaction_hash)
    return signature.vrs



from collections.abc import (
    Mapping,
)
from typing import (
    Union,
    Optional,
    Any,
    cast,
    Dict,
)

from eth_utils.curried import (
    text_if_str,
    to_bytes,
    to_text,
)
from hexbytes import (
    HexBytes,
)

from cfx_account._utils.structured_data.eth_account_legacy_hashing import (
    hash_cip23_message,
)

from eth_account.messages import (
    SignableMessage
)
from cfx_account._utils.structured_data.hashing import (
    hash_domain,
    load_and_validate_structured_message,
)
from cfx_account._utils.structured_data.validation import (
    validate_structured_data,
)

text_to_bytes = text_if_str(to_bytes)


def encode_structured_data(
    primitive: Optional[Union[bytes, int, Mapping]] = None, # type: ignore
    *,
    hexstr: Optional[str] = None,
    text: Optional[str] = None) -> SignableMessage:
    """
    Encode an CIP-23_ message.

    CIP-23 is the "structured data" approach.

    Supply the message as exactly one of the three arguments:

        - primitive, as a dict that defines the structured data
        - primitive, as bytes
        - text, as a json-encoded string
        - hexstr, as a hex-encoded (json-encoded) string

    :param primitive: the binary message to be signed
    :type primitive: bytes or int or Mapping (eg~ dict )
    :param hexstr: the message encoded as hex
    :param text: the message as a series of unicode characters (a normal Py3 str)
    :returns: The CIP-23 encoded message for typed data, ready for signing

    .. _CIP-23: https://github.com/Conflux-Chain/CIPs/blob/master/CIPs/cip-23.md
    """
    if isinstance(primitive, Mapping):
        validate_structured_data(cast(Dict[str, Any], primitive))
        structured_data = primitive
    else:
        message_string = to_text(primitive, hexstr=hexstr, text=text)
        structured_data = load_and_validate_structured_message(message_string)
    return SignableMessage(
        HexBytes(b'\x01'),
        hash_domain(structured_data), # type: ignore
        hash_cip23_message(structured_data),
    )
    
def encode_defunct(
        primitive: Optional[bytes] = None,
        *,
        hexstr: Optional[str] = None,
        text: Optional[str] = None) -> SignableMessage:
    r"""
    Encode a message for signing, using an old, unrecommended approach.
    :param primitive: the binary message to be signed
    :type primitive: bytes or int
    :param str hexstr: the message encoded as hex
    :param str text: the message as a series of unicode characters (a normal Py3 str)
    :returns: The CIP-23 encoded message for a string, ready for signing
    """
    message_bytes = to_bytes(primitive, hexstr=hexstr, text=text)
    msg_length = str(len(message_bytes)).encode('utf-8')

    # Encoding version E is defined by EIP-191 
    # "\x19Ethereum Signed Message:\n" + len(message).
    # There is not a similar definition in CIP-23, so we use C as version to compute, 
    # in which way the computing result is right
    return SignableMessage(
        b'C',
        b'onflux Signed Message:\n' + msg_length,
        message_bytes,
    )



from cfx_account import Account
from tests.test_utils import assert_hex_equal

private_key = "0x0123456789abcdef0123456789abcdef0123456789abcdef0123456789abcdef"
keystore = {
    "version": 3,
    "id": "db029583-f1bd-41cc-aeb5-b2ed5b33227b",
    "address": "1cad0b19bb29d4674531d6f115237e16afce377c",
    "crypto": {
        "ciphertext": "3198706577b0880234ecbb5233012a8ca0495bf2cfa2e45121b4f09434187aba",
        "cipherparams": {"iv": "a9a1f9565fd9831e669e8a9a0ec68818"},
        "cipher": "aes-128-ctr",
        "kdf": "scrypt",
        "kdfparams": {
            "dklen": 32,
            "salt": "3ce2d51bed702f2f31545be66fa73d1467d24686059776430df9508407b74231",
            "n": 8192,
            "r": 8,
            "p": 1,
        },
        "mac": "cf73832f328f3d5d1e0ec7b0f9c220facf951e8bba86c9f26e706d2df1e34890",
    },
}


def test_decrypt():
    assert_hex_equal(Account.decrypt(keystore, "password"), private_key)

def test_encrypt():
    encrypted = Account.encrypt(private_key, "password")
    assert_hex_equal(Account.decrypt(encrypted, "password"), private_key)



from cfx_account import Account
from tests.test_utils import assert_hex_equal

mnemonic = 'faint also eye industry survey unhappy boil public lemon myself cube sense'
private_key_0 = "0x40d0f137665463584cc57fce2b761572a85d1cbf1601fc93d001c129b2a11c92"

def test_generate_from_mnemonic():
    acct = Account.from_mnemonic(mnemonic, network_id=1)
    assert_hex_equal(acct.key, private_key_0)
    assert acct.address == 'cfxtest:aargrnff46pmuy2g1mmrntctkhr5mzamh6nmg361n0'
    
def test_account_restore():
    a1, mnemonic = Account.create_with_mnemonic(num_words=24, passphrase="TESTING", network_id=1)
    a2 = Account.from_mnemonic(mnemonic, passphrase="TESTING")
    assert a1.address.hex_address == a2.address # type: ignore
    



import pytest
from cfx_account.account import (
    Account
)
from cfx_utils.exceptions import (
    InvalidNetworkId
)
from cfx_utils.types import (
    Drip,
    CFX
)
from hexbytes import (
    HexBytes,
)
from .test_utils import assert_hex_equal
# from conflux_web3.dev import get_testnet_web3

key = '0xcc7939276283a32f60d2fad7d16cac972300308fe99ec98d0e63765d02e24863'
address = '0x1b981f81568eDD843DcB5b407ff0DD2e25618622'
base32_address = 'cfxtest:aar3uh6bm4hr5bb73rrya99u5y1cm2pgeja196rfeb'
expected_raw_tx = '0xf861dd0101649413d2ba4ed43542e7c54fbb6c5fccb9f269c1f94c016464018080a0a52f639cbed11262a7b88d0a37aef909aa7dc2c36c40689a3d52b8bd1d9482dea054f3bdeb654f73704db4cbc12451fb4c9830ef62b0f24de1a40e4b6fe10f57b2'
v = 0
r = '0xa52f639cbed11262a7b88d0a37aef909aa7dc2c36c40689a3d52b8bd1d9482de'
s = '0x54f3bdeb654f73704db4cbc12451fb4c9830ef62b0f24de1a40e4b6fe10f57b2'
signed_tx_hash = '0x692a0ea530a264f4e80ce39f393233e90638ef929c8706802e15299fd0b042b9'

transaction = {
    # 'from': '0x1b981f81568edd843dcb5b407ff0dd2e25618622'.lower(),
    'from': base32_address,
    'to': 'cfxtest:aak7fsws4u4yf38fk870218p1h3gxut3ku00u1k1da',
    'nonce': 1,
    'value': 1,
    'gas': 100,
    'gasPrice': 1,
    'storageLimit': 100,
    'epochHeight': 100,
    'chainId': 1
}

transaction_value_in_token = {
    # 'from': '0x1b981f81568edd843dcb5b407ff0dd2e25618622'.lower(),
    'from': base32_address,
    'to': 'cfxtest:aak7fsws4u4yf38fk870218p1h3gxut3ku00u1k1da',
    'nonce': 1,
    'value': CFX(1)/10**18,
    'gas': 100,
    'gasPrice': Drip(1),
    'storageLimit': 100,
    'epochHeight': 100,
    'chainId': 1
}

def test_sign_and_recover():
    account = Account.from_key(key)
    assert account.address == address
    signed_tx = Account.sign_transaction(transaction, key)
    # signed_tx = account.sign_transaction(transaction)
    assert v == signed_tx.v
    assert_hex_equal(r, signed_tx.r)
    assert_hex_equal(s, signed_tx.s)
    assert_hex_equal(signed_tx_hash, signed_tx.hash)
    assert_hex_equal(expected_raw_tx, signed_tx.raw_transaction)
    recovered_address = Account.recover_transaction(expected_raw_tx)
    assert recovered_address == address

def test_sign_and_recover_in_token():
    transaction = transaction_value_in_token
    account = Account.from_key(key)
    assert account.address == address
    signed_tx = Account.sign_transaction(transaction, key)
    assert v == signed_tx.v
    assert_hex_equal(r, signed_tx.r)
    assert_hex_equal(s, signed_tx.s)
    assert_hex_equal(signed_tx_hash, signed_tx.hash)
    assert_hex_equal(expected_raw_tx, signed_tx.raw_transaction)
    recovered_address = Account.recover_transaction(expected_raw_tx)
    assert recovered_address == address

def test_local_account():
    assert Account.from_key(key).address == address
    assert Account.from_key(key, network_id=1).address == base32_address
    assert Account.from_key(key, network_id=1).hex_address == address
    assert Account.from_key(key).get_base32_address(1) == base32_address
    
    local_account = Account.from_key(key)
    local_account.network_id = 1
    assert local_account.network_id == 1
    with pytest.raises(InvalidNetworkId):
        local_account.network_id = 0
    assert local_account.address == base32_address
    local_account.network_id = None
    assert local_account.address == address

# def test_set_network_id():
#     assert Account.create().address.startswith("0x")
#     assert Account.create(network_id=1029).network_id == 1029
#     w3 = get_testnet_web3()
#     Account.set_w3(w3)
#     assert Account.create().network_id == 1
#     Account.set_w3(None) # type: ignore



from typing import Any, ClassVar, Dict, Tuple

import rlp
from eth_account._utils.transaction_utils import transaction_rpc_to_rlp_structure
from eth_rlp import HashableRLP
from eth_utils import keccak
from eth_utils.curried import apply_formatters_to_dict
from hexbytes import HexBytes
from rlp.sedes import Binary, big_endian_int, binary
from toolz import dissoc, merge, partial, pipe
from typing_extensions import Self

from cfx_account.transactions.base import TransactionImplementation
from cfx_account.transactions.transaction_utils import access_list_sede_type

from .transaction_utils import TYPED_TRANSACTION_FORMATTERS


class CIP1559Transaction(TransactionImplementation):

    transaction_type: ClassVar[int] = 2

    unsigned_transaction_fields = (
        ("nonce", big_endian_int),
        ("maxPriorityFeePerGas", big_endian_int),
        ("maxFeePerGas", big_endian_int),
        ("gas", big_endian_int),
        ("to", Binary.fixed_length(20, allow_empty=True)),
        ("value", big_endian_int),
        ("storageLimit", big_endian_int),
        ("epochHeight", big_endian_int),
        ("chainId", big_endian_int),
        ("data", binary),
        ("accessList", access_list_sede_type),
    )

    signature_fields = (
        ("v", big_endian_int),
        ("r", big_endian_int),
        ("s", big_endian_int),
    )

    transaction_field_defaults = {
        # "type": b"0x2",
        "chainId": 0,
        "to": b"",
        "value": 0,
        "data": b"",
        "accessList": [],
    }

    _unsigned_transaction_serializer = type(
        "_unsigned_transaction_serializer",
        (HashableRLP,),
        {
            "fields": unsigned_transaction_fields,
        },
    )

    signed_fields = (
        ("tx_meta", _unsigned_transaction_serializer),
        ("v", big_endian_int),
        ("r", big_endian_int),
        ("s", big_endian_int),
    )

    _signed_transaction_serializer = type(
        "_signed_transaction_serializer",
        (HashableRLP,),
        {"fields": signed_fields},
    )

    # Noting that the transaction validity is not checked here
    def __init__(self, transaction_dict: Dict[str, Any]):
        if "type" in transaction_dict:
            transaction_dict.pop("type")
        self.__class__.ensure_no_fields_missing(transaction_dict)
        self.__class__.ensure_no_fields_extra(transaction_dict)

        sanitized_dictionary = pipe(
            transaction_dict,
            dict,
            partial(merge, self.__class__.transaction_field_defaults),
            apply_formatters_to_dict(TYPED_TRANSACTION_FORMATTERS),
        )

        self._dictionary = sanitized_dictionary

    def hash(self) -> bytes:
        """
        Noting: this is not the transaction hash
        Hashes this DynamicFeeTransaction to prepare it for signing.
        As per the CIP-1559 specifications, the signature is a secp256k1 signature over
        ``keccak256(b'cfx' || 0x02 || rlp([nonce, maxPriorityFeePerGas,
        maxFeePerGas, gas, to, value, data, storageLimit, epochHeight, accessList]))``
        """
        return keccak(self._encode_unsigned())

    def as_dict(self) -> Dict[str, Any]:
        return self._dictionary

    def vrs(self) -> Tuple[int, int, int]:
        """Returns (v, r, s) if they exist."""
        if not self.is_signed():
            raise ValueError("attempting to encode an unsigned transaction")
        return (self._dictionary["v"], self._dictionary["r"], self._dictionary["s"])

    def is_signed(self) -> bool:
        return "v" in self._dictionary

    def append_signature(self, *, v: int, r: int, s: int) -> Self:
        if self.is_signed():
            raise ValueError("attempting to sign a signed transaction")
        self._dictionary.update({"v": v, "r": r, "s": s})
        return self
    
    def _encode_unsigned(self) -> bytes:
        transaction_without_signature_fields = dissoc(self._dictionary, "v", "r", "s")
        rlp_structured_txn_without_sig_fields = transaction_rpc_to_rlp_structure(
            transaction_without_signature_fields
        )
        rlp_serializer = self.__class__._unsigned_transaction_serializer
        return pipe(
            rlp_serializer.from_dict(rlp_structured_txn_without_sig_fields),  # type: ignore  # noqa: E501
            lambda val: rlp.encode(val),  # type: ignore
            # (b'cfx' || 0x02 || rlp([...]))
            lambda val: HexBytes(b"cfx") + HexBytes("0x02") + HexBytes(val),  # type: ignore
        )

    def encode(self, *, allow_unsigned: bool = False) -> bytes:
        """
        Returns this raw transaction as bytes.

        The raw transaction is
            b'cfx' || 0x02 || TransactionPayload

        The transaction payload is:

            TransactionPayload = rlp([ nonce, maxPriorityFeePerGas,
            maxFeePerGas, gasLimit, to, value, storageLimit, epochHeight
            data, accessList, signatureYParity, signatureR, signatureS])
        """
        if not self.is_signed():
            if not allow_unsigned:
                raise ValueError("attempting to encode an unsigned transaction without allow_unsigned=True")
            return self._encode_unsigned()
        rlp_serializer = self.__class__._signed_transaction_serializer
        rlp_structured_dict = transaction_rpc_to_rlp_structure(self._dictionary)
        tx_meta = dissoc(rlp_structured_dict, "v", "r", "s")
        payload = rlp.encode(
            rlp_serializer(
                tx_meta=self._unsigned_transaction_serializer(**tx_meta),
                v=self._dictionary["v"],
                r=self._dictionary["r"],
                s=self._dictionary["s"],
            )
        )
        return HexBytes(b"cfx") + HexBytes("0x02") + payload

    @classmethod
    def ensure_no_fields_missing(cls, dictionary: Dict[str, Any]):
        # add default fields
        for fields in cls.transaction_field_defaults:
            if fields not in dictionary:
                dictionary[fields] = cls.transaction_field_defaults[fields]
        # check unsigned fields
        for field, _ in cls.unsigned_transaction_fields:
            if field not in dictionary:
                raise ValueError(f"{field} is missing in {dictionary}")
        # if this is a dictionary with signed fields
        if "v" in dictionary or "r" in dictionary or "s" in dictionary:
            for field, _ in cls.signature_fields:
                if field not in dictionary:
                    raise ValueError(f"{field} is missing in {dictionary}")

    @classmethod
    def ensure_no_fields_extra(cls, dictionary: Dict[str, Any]):
        expected_fields = [
            field
            for field, _ in (cls.unsigned_transaction_fields + cls.signature_fields)
        ]
        for field in dictionary:
            if field not in expected_fields:
                raise ValueError(
                    f"{field} is not a valid field: expecting fields - {expected_fields}"
                )



from typing import (
    Union,
)
from typing_extensions import (
    Literal,
    TypedDict,
)

from eth_typing import (
    HexStr
)

class ScryptParams(TypedDict):
    dklen: int
    n: int # 262144 is default factor
    r: int
    p: int
    salt: HexStr

class PBKDF2Params(TypedDict):
    c: int # 1000000 is default factor
    dklen: int
    prf: Literal['hmac-sha256']
    salt: HexStr

class CipherParams(TypedDict):
    iv: HexStr

class CryptoParams(TypedDict):
    cipher: Literal['aes-128-ctr']
    cipherparams: CipherParams
    ciphertext: HexStr
    kdf: Literal["scrypt", "pbkdf2"]
    kdfparams: Union[PBKDF2Params, ScryptParams]
    mac: HexStr

class KeyfileDict(TypedDict):
    address: HexStr
    crypto: CryptoParams
    id: HexStr
    version: Literal[3]



from typing import (
    Dict,
    Any,
)
import json
from eth_utils import (
    keccak,
)
from eth_account._utils.encode_typed_data.encoding_and_hashing import (
    encode_data
)

from cfx_account._utils.structured_data.validation import (
    validate_structured_data
)

def load_and_validate_structured_message(structured_json_string_data: str) -> Dict[str, Any]:
    structured_data = json.loads(structured_json_string_data)
    validate_structured_data(structured_data)

    return structured_data

def hash_domain(structured_data: Dict[str, Any]) -> bytes:
    return keccak(
        encode_data(
            "CIP23Domain",
            structured_data["types"],
            structured_data["domain"]
        )
    )


from typing import Type

from cfx_utils.types import TxParam
from hexbytes import HexBytes

from .base import TransactionImplementation
from .cip1559_transactions import CIP1559Transaction
from .legacy_transactions import LegacyTransaction
from .transaction_utils import copy_ensuring_int_transaction_type


class Transaction:
    """
    Represents Conflux Transaction including legacy transaction or typed transactions.
    The currently supported Transaction Types are:

     * type 0: LegacyTransaction
     * type 1: CIP-2930's AccessListTransaction
     * type 2: CIP-1559's DynamicFeeTransaction

    """

    transaction_impl: TransactionImplementation

    # def __init__(self, transaction_type: int, transaction: TransactionImplementation):
    #     """Should not be called directly. Use instead the 'from_dict' method."""
    #     if not isinstance(transaction, TransactionImplementation):
    #         raise TypeError(
    #             f"expected _TransactionImplementation, got {type(transaction)}"
    #         )
    #     if not transaction_type == transaction.transaction_type:
    #         raise ValueError(
    #             f"expected transaction type {repr(transaction_type)}, got {repr(transaction.transaction_type)}"
    #         )
    #     self.transaction = transaction

    # @property
    # def blob_data(self) -> Optional[BlobPooledTransactionData]:
    #     """Returns the blobs associated with this transaction."""
    #     return self.transaction.blob_data

    @classmethod
    def from_dict(
        cls,
        dictionary: TxParam,  # blobs: List[bytes] = None
    ) -> "TransactionImplementation":
        """
        Builds a TypedTransaction from a dictionary.
        Verifies the dictionary is well formed.
        """
        dict_copy = copy_ensuring_int_transaction_type(dictionary)
        # if not is_int_or_prefixed_hexstr(dict_copy["type"]):
        #     raise ValueError("incorrect transaction type")
        # Switch on the transaction type to choose the correct constructor.
        transaction_type: int = dict_copy["type"]
        transaction: Type[TransactionImplementation]
        if transaction_type == LegacyTransaction.transaction_type:
            transaction = LegacyTransaction
        elif transaction_type == CIP1559Transaction.transaction_type:
            transaction = CIP1559Transaction
        # elif transaction_type == AccessListTransaction.transaction_type:
        #     transaction = AccessListTransaction
        # elif transaction_type == BlobTransaction.transaction_type:
        #     transaction = BlobTransaction
        else:
            raise TypeError(f"Unknown Transaction type: {transaction_type}")
        return transaction(dict_copy)

    @classmethod
    def from_bytes(cls, encoded_transaction: HexBytes) -> "TransactionImplementation":
        """Builds a TypedTransaction from a signed encoded transaction."""

        if not isinstance(encoded_transaction, HexBytes):
            raise TypeError(f"expected Hexbytes, got {type(encoded_transaction)}")
        return LegacyTransaction.from_bytes(encoded_transaction)

        # if not (len(encoded_transaction) > 0 and encoded_transaction[0] <= 0x7F):
        #     raise ValueError("unexpected input")

        # transaction: Union[
        #     "DynamicFeeTransaction", "AccessListTransaction", "BlobTransaction"
        # ]

        # if encoded_transaction[0] == AccessListTransaction.transaction_type:
        #     transaction_type = AccessListTransaction.transaction_type
        #     transaction = AccessListTransaction.from_bytes(encoded_transaction)
        # elif encoded_transaction[0] == DynamicFeeTransaction.transaction_type:
        #     transaction_type = DynamicFeeTransaction.transaction_type
        #     transaction = DynamicFeeTransaction.from_bytes(encoded_transaction)
        # elif encoded_transaction[0] == BlobTransaction.transaction_type:
        #     transaction_type = BlobTransaction.transaction_type
        #     transaction = BlobTransaction.from_bytes(encoded_transaction)
        # else:
        #     # The only known transaction types should be explicit if/elif branches.
        #     raise TypeError(
        #         f"typed transaction has unknown type: {encoded_transaction[0]}"
        #     )
        # return cls(
        #     transaction_type=transaction_type,
        #     transaction=transaction,
        # )
        raise NotImplementedError

    # def hash(self) -> bytes:
    #     """
    #     Hashes this transaction to prepare it for signing,
    #     noting this is not the hash of the raw transaction as the signature is not included.
    #     """
    #     return self.unsigned_hash()

    # def unsigned_hash(self):
    #     """
    #     Hashes this transaction to prepare it for signing,
    #     noting this is not the hash of the raw transaction as the signature is not included.
    #     """
    #     return self.transaction.hash()

    # def append_signature(self, *, v: int, r: int, s: int):
    #     self.transaction.append_signature(v=v, r=r, s=s)
    #     return self

    # def encode(self) -> bytes:
    #     """
    #     Encodes this Transaction and returns it as bytes.
    #     Exception will be raised if signature is not attached.

    #     The transaction format follows EIP-2718's typed transaction
    #     format (TransactionType || TransactionPayload).
    #     Note that we delegate to a transaction type's payload() method as
    #     the EIP-2718 does not prescribe a TransactionPayload format,
    #     leaving types free to implement their own encoding.
    #     """
    #     if self.transaction_type != 0:
    #         return bytes([self.transaction_type]) + self.transaction.payload()
    #     return self.transaction.encode()

    # def as_dict(self) -> Dict[str, Any]:
    #     """Returns this transaction as a dictionary."""
    #     # return normalize_transaction_dict(self.transaction.as_dict())
    #     return self.transaction.as_dict()

    # def vrs(self) -> Tuple[int, int, int]:
    #     """Returns (v, r, s) if they exist."""
    #     return self.transaction.vrs()



from typing import TYPE_CHECKING, Any, Optional, Union, Type
from typing_extensions import Literal
from eth_utils.address import to_checksum_address
from eth_account.signers.local import LocalAccount as EthLocalAccount
from eth_account.datastructures import (
    SignedMessage,
    SignedTransaction,
)
from eth_account.messages import (
    SignableMessage
)
from cfx_address import (
    eth_eoa_address_to_cfx_hex,
    Base32Address
)
from cfx_address.utils import (
    validate_network_id
)
from cfx_utils.types import (
    TxParam,
    ChecksumAddress,
)
from cfx_account.types import (
    KeyfileDict,
)


if TYPE_CHECKING:
    from cfx_account import Account


class LocalAccount(EthLocalAccount):
    def __init__(self, key: Any, account: Union["Account", Type["Account"]], network_id: Optional[int]=None):
        if network_id is not None:
            validate_network_id(network_id)
        self._network_id = network_id
        
        super().__init__(key, account)

    @property
    def network_id(self) -> Union[int, None]:
        """
        The network id of the account, which determines its address.
        Can be set using a positive int or None

        :return Union[int, None]: the network id of the account, 1029 for mainnet and 1 for testnet
        
        :examples:
        
        >>> from cfx_account import Account
        >>> acct = Account.create()
        >>> assert acct.network_id is None
        >>> acct.address
        '0x1261D876E5C589Bc26248c53C68b71c2d28470e9'  
        >>> acct.network_id = 1
        >>> acct.address
        'cfxtest:aakgd0d061c2xtbgewgfhvynshbrfbdu7e55kdxfxx'
        """        
        return self._network_id
    
    @network_id.setter
    def network_id(self, new_network_id: Union[int, None]) -> None:
        """
        Set the network id of the account, which determines its address

        :param Union[int, None] new_network_id: the new network id
        """        
        if new_network_id is not None:
            validate_network_id(new_network_id)
        self._network_id = new_network_id
        
    # def reset_network_id(self):
    #     self._network_id = None

    @property
    def address(self) -> Union[Base32Address, ChecksumAddress]:
        """
        Returns the address of the account.

        :return Union[Base32Address,ChecksumAddress]: Returns a Base32Address if network id is not None, 
            else ChecksumAddress
        """
        hex_address = self.hex_address
        if not self._network_id:
            return hex_address
        return Base32Address(hex_address, self._network_id)

    @property
    def hex_address(self) -> ChecksumAddress:
        """
        Returns the hex address of the account.

        :return ChecksumAddress: the hex address in checksum format
        """        
        return to_checksum_address(eth_eoa_address_to_cfx_hex(super().address))
    
    @property
    def key(self) -> bytes:
        """
        Get the private key.
        """
        return super().key
    
    def get_base32_address(self, specific_network_id: int) -> Base32Address:
        """
        Returns the Base32Address of the account in specific network
        without changing the account network id

        :param int specific_network_id: the network id of the target network
        :return Base32Address:
        """        
        return Base32Address(self.address, specific_network_id)
    
    def sign_transaction(self, transaction_dict: TxParam) -> SignedTransaction:
        """
        This uses the same structure as in
        :meth:`~cfx_account.account.Account.sign_transaction`, but without a private key argument.
        """
        return super().sign_transaction(transaction_dict)
    
    def sign_message(self, signable_message: SignableMessage) -> SignedMessage:
        """
        This uses the same structure as in
        :meth:`~cfx_account.account.Account.sign_message`, but without a private key argument.
        """
        return super().sign_message(signable_message)
    
    def encrypt(self, password: str, kdf: Optional[Literal['scrypt', 'pbkdf2']]=None, iterations: Optional[int]=None) -> KeyfileDict:
        """
        This uses the same structure as in
        :meth:`~cfx_account.account.Account.encrypt`, but without a private key argument.
        """
        return super().encrypt(password, kdf, iterations)
    
    def __bytes__(self) -> bytes:
        return self.key
